from typing import List, Optional


class MaximCache():

    def get_all_keys(self) -> List[str]:
        return []

    def get(self, key: str) -> Optional[str]:
        pass

    def set(self, key: str, value: str) -> None:
        pass

    def delete(self, key: str) -> None:
        pass
